/*
/*
    File: SampleFormat.c

    Copyright 1993-1994 by Adobe Systems, Inc.

*/

#include "PIUtilities.h"
#include "FormatUtilities.h"

#if defined(THINK_C) || defined(__MWERKS__)

#define ENTRYPOINT main

#endif

#if Macintosh

#include "DialogUtilities.h"

#endif

/*****************************************************************************/

void InitGlobals (GPtr globals);

void DoReadPrepare (GPtr globals);
void DoReadStart (GPtr globals);
void DoReadContinue (GPtr globals);
void DoReadFinish (GPtr globals);
void DoOptionsPrepare (GPtr globals);
void DoOptionsStart (GPtr globals);
void DoOptionsContinue (GPtr globals);
void DoOptionsFinish (GPtr globals);
void DoEstimatePrepare (GPtr globals);
void DoEstimateStart (GPtr globals);
void DoEstimateContinue (GPtr globals);
void DoEstimateFinish (GPtr globals);
void DoWritePrepare (GPtr globals);
void DoWriteStart (GPtr globals);
void DoWriteContinue (GPtr globals);
void DoWriteFinish (GPtr globals);
void DoFilterFile (GPtr globals);

/*****************************************************************************/

typedef struct FileHeader
{ 
	unsigned char	widthH ;				// Byte 0 and 1 
	unsigned char	widthL ;
	unsigned char	heightH ;
	unsigned char	heightL ;
	unsigned char	xoffsetH ;
	unsigned char	xoffsetL ;
	unsigned char	yoffsetH ;
	unsigned char	yoffsetL ;
	unsigned char	bppH ;
	unsigned char	bppL ;
} FileHeader ; 
 
/*****************************************************************************/

/* Main dispatching routine.  Initializes and sets up the global variables,
   and performs the operation specified by the selector. */

#if MSWindows
void ENTRYPOINT (short selector,
						FormatRecordPtr stuff,
						long *data,
						short *result)
#else
pascal void ENTRYPOINT (short selector,
						FormatRecordPtr stuff,
						long *data,
						short *result)
#endif
	{

	/* Set up the globals. */

	Globals globalValues;
	GPtr globals = &globalValues;
	
	if (!*data)
		{
		
		InitGlobals (globals);

		*data = (long) NewHandle (sizeof (Globals));
		
		if (!*data)
			{
			*result = memFullErr;
			return;
			}
			
		** (GHdl) *data = globalValues;
		
		}
		
	globalValues = ** (GHdl) *data;
	
	/* Perform the requested operation */

	gStuff  = stuff;
	gResult = noErr;

	switch (selector)
		{

		case formatSelectorAbout:
			DoAbout (globals);
			break;

		case formatSelectorReadPrepare:
			DoReadPrepare (globals);
			break;

		case formatSelectorReadStart:
			DoReadStart (globals);
			break;

		case formatSelectorReadContinue:
			DoReadContinue (globals);
			break;

		case formatSelectorReadFinish:
			DoReadFinish (globals);
			break;

		case formatSelectorOptionsPrepare:
			DoOptionsPrepare (globals);
			break;

		case formatSelectorOptionsStart:
			DoOptionsStart (globals);
			break;

		case formatSelectorOptionsContinue:
			DoOptionsContinue (globals);
			break;

		case formatSelectorOptionsFinish:
			DoOptionsFinish (globals);
			break;

		case formatSelectorEstimatePrepare:
			DoEstimatePrepare (globals);
			break;

		case formatSelectorEstimateStart:
			DoEstimateStart (globals);
			break;

		case formatSelectorEstimateContinue:
			DoEstimateContinue (globals);
			break;

		case formatSelectorEstimateFinish:
			DoEstimateFinish (globals);
			break;

		case formatSelectorWritePrepare:
			DoWritePrepare (globals);
			break;

		case formatSelectorWriteStart:
			DoWriteStart (globals);
			break;

		case formatSelectorWriteContinue:
			DoWriteContinue (globals);
			break;

		case formatSelectorWriteFinish:
			DoWriteFinish (globals);
			break;
			
		case formatSelectorFilterFile:
			DoFilterFile (globals);
			break;

		default:
			gResult = formatBadParameters;
		}

	*result = gResult;

	** (GHdl) *data = globalValues;
		
	}

/*****************************************************************************/

/* Sets the global variables to their default values. */

void InitGlobals (GPtr globals)
	{

	Ptr		p;
	int16	count;
	
	p = (Ptr) globals;
	count = sizeof(Globals);
	
	while (count--)
		*p++ = 0;

	}

/*****************************************************************************/

static Boolean CheckForServices (GPtr globals)
	{
	
	return WarnBufferProcsAvailable () &&
		   WarnAdvanceStateAvailable () &&
		   WarnHandleProcsAvailable ();
	
	}

/*****************************************************************************/

static Boolean TestAndStoreResult (GPtr globals, OSErr result)
	{
	
	if (result != noErr)
		{
		if (gResult == noErr)
			gResult = result;
		return FALSE;
		}
	else
		return TRUE;
	
	}
	
#define TSR(x) TestAndStoreResult (globals, x)

/*****************************************************************************/

static Boolean TestAndStoreCancel (GPtr globals, Boolean cancel)
	{
	
	if (cancel)
		{
		if (gResult == noErr)
			gResult = userCanceledErr;
		return FALSE;
		}
	else
		return TRUE;
	
	}
	
#define TSC(x) TestAndStoreCancel (globals, x)

/*****************************************************************************/

static int32 RowBytes (GPtr globals)
	{
	
	return (gStuff->imageSize.h * (int32) gStuff->depth + 7) >> 3;
	
	}

/*****************************************************************************/

static Boolean AllocatePixelBuffer (GPtr globals)
	{
	
	BufferID buffer;
	
	if (gResult != noErr) return FALSE;

	/* We will want a buffer that is one line wide. */
	
	gPixelBuffer = 0;
	
	gRowBytes = RowBytes (globals);
	
	if (!TSR (AllocateBuffer (gRowBytes * 4 * 2, &buffer))) return FALSE;
	
	gPixelBuffer = buffer;
	
	gPixelData = LockBuffer (gPixelBuffer, FALSE);
	
	return TRUE;
	
	}

/*****************************************************************************/

static void DisposePixelBuffer (GPtr globals)
	{
	
	if (gPixelBuffer != 0)
		{
		
		FreeBuffer (gPixelBuffer);
		
		gPixelBuffer = 0;
		gPixelData = 0;
		
		}
	
	}

/*****************************************************************************/

void DoReadPrepare (GPtr globals)
	{
	
	gStuff->maxData = 0;
	
	}

/*****************************************************************************/

static void ReadSome (GPtr globals, int32 count, void *buffer)
	{
	
	int32 readCount = count;
	
	if (gResult != noErr)
		return;
	
	gResult = FSRead (gStuff->dataFork, &readCount, buffer);
	
	if (gResult == noErr && readCount != count)
		gResult = eofErr;
	
	}

/*****************************************************************************/

static void WriteSome (GPtr globals, int32 count, void *buffer)
	{
	
	int32 writeCount = count;
	
	if (gResult != noErr)
		return;
	
	gResult = FSWrite (gStuff->dataFork, &writeCount, buffer);
	
	if (gResult == noErr && writeCount != count)
		gResult = dskFulErr;
	
	}

/*****************************************************************************/

static void ReadRow (GPtr globals)
	{
	
	ReadSome (globals, gRowBytes, gPixelData);
	
	}

/*****************************************************************************/

static void WriteRow (GPtr globals)
	{
	
	WriteSome (globals, gRowBytes, gPixelData);
	
	}

/*****************************************************************************/

static void DisposeImageResources (GPtr globals)
	{
	
	if (gStuff->imageRsrcData)
		{
		
		PIDisposeHandle (gStuff->imageRsrcData);
		
		gStuff->imageRsrcData = NULL;
		
		gStuff->imageRsrcSize = 0;
		
		}
	
	}

/*****************************************************************************/

void DoReadStart (GPtr globals)
	{
	
	FileHeader header;
	
	/* Exit if we have already encountered an error. */

    if (gResult != noErr) return;
		
	/* Check for the needed services. */
	
	if (!TSC (!CheckForServices (globals))) return;

	/* If we have not encountered an error, then we want to read
	   the file header. */
	   
	if (!TSR (SetFPos (gStuff->dataFork, fsFromStart, 0))) return;
	
	ReadSome (globals, sizeof (FileHeader), &header);
	
	if (gResult != noErr) return;

	if (header.bppL != 8 && header.bppL != 24)
	{
		gResult = formatCannotRead;
		return;
	}
	
	gStuff->imageMode   = (header.bppL == 24) ? plugInModeRGBColor : plugInModeGrayScale;
	gStuff->imageSize.h = header.widthH  * 256 + header.widthL;
	gStuff->imageSize.v = header.heightH * 256 + header.heightL;
	gStuff->depth       = 8;
	gStuff->planes      = (header.bppL == 24) ? 3 : 1;
	
	return;
		
	}

/*****************************************************************************/

void mymemmove (void *dst, void *src, size_t size)
{
	unsigned char *s, *d;

	s = (unsigned char *)src;
	d = (unsigned char *)dst;

	while (size)
	{
		*d++ = *s++;
		size--;
	}
}

/*****************************************************************************/

void DoReadContinue (GPtr globals)
	{
	
	int32 done;
	int32 total;
	int16 row;
	
	/* Set up the progress variables. */
	
	done = 0;
	total = gStuff->imageSize.v;
		
	/* Next, we will allocate the pixel buffer. */
	
	AllocatePixelBuffer (globals);
	
	/* Set up to start returning chunks of data. */
	
	gStuff->theRect.left  = 0;
	gStuff->theRect.right = gStuff->imageSize.h;
	gStuff->loPlane    = 0;
	gStuff->hiPlane    = (gStuff->imageMode == plugInModeRGBColor) ? 2 : 0;
	gStuff->colBytes   = (gStuff->imageMode == plugInModeRGBColor) ? 3 : 1;
	gStuff->rowBytes   = (gStuff->imageMode == plugInModeRGBColor) ?
		(gRowBytes * 3) : 
		(gRowBytes    ) ;
	gStuff->planeBytes = 1;

	gStuff->data = gPixelData;

	{
		unsigned char	*loadBuffer;
		unsigned char	*pixelBuffer;
		long			 bytesInBuffer;
		long			 bufferSize;

		bytesInBuffer = 0;
		bufferSize    = gRowBytes * 4;
		loadBuffer    = (unsigned char *)gPixelData + bufferSize;
		pixelBuffer   = (unsigned char *)gPixelData;

		for (row = 0; gResult == noErr && row < gStuff->imageSize.v; ++row)
		{
			unsigned char *pixels;
			unsigned char *data;
			int32		readWidth = 0;

			pixels = pixelBuffer;
			data   = loadBuffer;
			
			gStuff->theRect.top    = row;
			gStuff->theRect.bottom = row + 1;
		
			if (bytesInBuffer)
			{
				mymemmove (loadBuffer, loadBuffer + bufferSize - bytesInBuffer, bytesInBuffer);
			}
			ReadSome (globals, bufferSize - bytesInBuffer, loadBuffer + bytesInBuffer);
			bytesInBuffer = bufferSize;

			if (gResult == noErr || gResult == eofErr)
			{
				int	count;
				int width;
				unsigned char r,g,b;

				width = gStuff->imageSize.h;

				if (gStuff->imageMode == plugInModeRGBColor)
				{
	  				while (width > 0)
					{
   						count = *data++;
						b     = *data++;
   						g     = *data++;
						r     = *data++;

						bytesInBuffer -= 4;

						width -= count;
						if (width < 0)
						{
							gResult = readErr;
							goto cleanup;
						}

						if (count == 0)
						{
							gResult = readErr;
							goto cleanup;
						}

						while (count)
						{
							*pixels++ = r;
							*pixels++ = g;
							*pixels++ = b;
							count--;
						}
					}
				}
				else
				{
	  				while (width > 0)
					{
   						count = *data++;
						b     = *data++;

						bytesInBuffer -= 2;

						width -= count;
						if (width < 0)
						{
							gResult = readErr;
							goto cleanup;
						}

						if (count == 0)
						{
							gResult = readErr;
							goto cleanup;
						}

						while (count)
						{
							*pixels++ = b;
							count--;
						}
					}
				}

				if (width < 0)
				{
					gResult = readErr;
					goto cleanup;
				}

				gResult = AdvanceState ();
			}
				
			UpdateProgress (row, total);
			
		}
	}

cleanup:
	gStuff->data = NULL;
	
	DisposePixelBuffer (globals);
	
	}

/*****************************************************************************/

void DoReadFinish (GPtr globals)
	{
	
	/* Dispose of the image resource data if it exists. */
	
	DisposeImageResources (globals);
	
	}

/*****************************************************************************/

void DoOptionsPrepare (GPtr globals)
	{

	gStuff->maxData = 0;

	}

/*****************************************************************************/

void DoOptionsStart (GPtr globals)
	{

	/* Check for the needed services. */
	
	if (!TSC (!CheckForServices (globals))) return;

	gStuff->data = NULL;

	}

/*****************************************************************************/

void DoOptionsContinue (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoOptionsFinish (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoEstimatePrepare (GPtr globals)
	{

	gStuff->maxData = 0;

	}

/*****************************************************************************/

void DoEstimateStart (GPtr globals)
	{
	
	int32 dataBytes;
	
	/* Check for the needed services. */
	
	if (!TSC (!CheckForServices (globals))) return;

	dataBytes = sizeof (FileHeader) +
				RowBytes (globals) * gStuff->planes * gStuff->imageSize.v;
					  
	if (gStuff->imageMode == plugInModeIndexedColor)
		dataBytes += 3 * sizeof (LookUpTable);
		
	gStuff->minDataBytes = dataBytes;
	gStuff->maxDataBytes = dataBytes;
	
	gStuff->data = NULL;

	}

/*****************************************************************************/

void DoEstimateContinue (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoEstimateFinish (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoWritePrepare (GPtr globals)
	{
	
	gStuff->maxData = 0;
	
	}

/*****************************************************************************/

void DoWriteStart (GPtr globals)
	{

	FileHeader header;
	int32 total;
	int16 row;
	
    if (gResult != noErr) return;
		
	/* Check for the needed services. */
	
	if (!TSC (!CheckForServices (globals))) return;

	/* Write the header. */
	
#if 0
	if (gStuff->imageMode != plugInModeRGBColor &&
		gStuff->imageMode != plugInModeGrayScale)
	{
		gResult = exportBadMode;
		return;
	}
#endif

	gResult = SetFPos (gStuff->dataFork, fsFromStart, 0);

	header.widthH = gStuff->imageSize.h / 256;
	header.widthL = gStuff->imageSize.h % 256;
	header.heightH = gStuff->imageSize.v / 256;
	header.heightL = gStuff->imageSize.v % 256;
	header.xoffsetH = 0;
	header.xoffsetL = 0;
	header.yoffsetH = (gStuff->imageSize.v - 1) / 256;
	header.yoffsetL = (gStuff->imageSize.v - 1) % 256;
	header.bppH = 0;
	header.bppL = (gStuff->imageMode == plugInModeRGBColor) ? 24 : 8; 
	
	WriteSome (globals, sizeof (FileHeader), &header);
	
	if (gResult != noErr) return;
	
	/* Set up the progress variables. */

	total = gStuff->imageSize.v;
		
	/* Next, we will allocate the pixel buffer. */
	
	AllocatePixelBuffer (globals);
		
	/* Set up to start receiving chunks of data. */
	
	gStuff->theRect.left  = 0;
	gStuff->theRect.right = gStuff->imageSize.h;
	gStuff->loPlane    = 0;
	gStuff->hiPlane    = (gStuff->imageMode == plugInModeRGBColor) ? 2 : 0;
	gStuff->colBytes   = (gStuff->imageMode == plugInModeRGBColor) ? 3 : 1;
	gStuff->rowBytes   = (gStuff->imageMode == plugInModeRGBColor) ?
		(gRowBytes * 3) : 
		(gRowBytes    ) ;
	gStuff->planeBytes = 1;

	gStuff->data = gPixelData;
	
	for (row = 0; gResult == noErr && row < gStuff->imageSize.v; ++row)
	{
		unsigned char *packBuffer;
		unsigned char *dst;
		unsigned char *src;
		int32	       bytesWritten = 0;
		int32		   width;

		packBuffer = (unsigned char *)gPixelData + gRowBytes * 4;
		dst        = packBuffer;
		src        = gPixelData;
 
		gStuff->theRect.top    = row;
		gStuff->theRect.bottom = row + 1;
			
		if (gResult == noErr || gResult == eofErr)
			gResult = AdvanceState ();

		width = gStuff->imageSize.h;
		while (width)
		{
			unsigned char r,g,b;
			int count;

			if (gStuff->imageMode == plugInModeRGBColor)
			{
				r = *src++;
				g = *src++;
				b = *src++;

				count = 1;
				width--;

				while (width && count < 255)
				{
					if (src[0] != r || src[1] != g || src[2] != b)
					{
						break;
					}
					src += 3;
					count++;
					width--;
				}

				*dst++ = (unsigned char)count;
				*dst++ = b;
				*dst++ = g;
				*dst++ = r;

				bytesWritten += 4;
			}
			else
			{
				b = *src++;

				count = 1;
				width--;

				while (width && count < 255)
				{
					if (src[0] != b)
					{
						break;
					}
					src += 1;
					count++;
					width--;
				}

				*dst++ = (unsigned char)count;
				*dst++ = b;

				bytesWritten += 2;
			}
		}
				
		WriteSome (globals, bytesWritten, packBuffer);
			
		UpdateProgress (row, total);
		
	}
		
	gStuff->data = NULL;
	
	DisposePixelBuffer (globals);

	}

/*****************************************************************************/

void DoWriteContinue (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoWriteFinish (GPtr globals)
	{
#ifndef __PPCC__
	#pragma unused (globals)
#endif
	}

/*****************************************************************************/

void DoFilterFile (GPtr globals)
	{
	
	FileHeader header;
	
	/* Exit if we have already encountered an error. */

    if (gResult != noErr) return;
		
	/* Read the file header. */
	   
	if (!TSR (SetFPos (gStuff->dataFork, fsFromStart, 0))) return;
	
	ReadSome (globals, sizeof (FileHeader), &header);
	
	if (gResult != noErr) return;
	
	/* Check what you can */
	if (header.bppL != 8 && header.bppL != 24)
	{
		gResult = formatCannotRead;
		return;
	}
	}

/*****************************************************************************/







